# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the hveto python package.
#
# hveto is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# hveto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with hveto.  If not, see <http://www.gnu.org/licenses/>.

"""
Configuration files for Hveto
#############################

How to write a configuration file
=================================

The `hveto` command-line executable can run without a configuration file, and wi
ll auto-select the primary channel, and attempt to auto-detect which auxiliary c
hannels are available. It will also inset a minimal list of 'unsafe' channels so that the analysis isn't totally meaningless.

However, the way that things are auto-detected means that the resulting output i
sn't exactly reproducable after-the-fact in general. If someone adds or removes data for an auxiliary channel, a rerun will operate over a different channel lis
ts, and confusion will reign.

So, it is highly recommended that you use a custom configuration file for your a
nalyses. It could be that the configuration file never changes, but the very fac
t that it exists should mean that you are in much greater control over what the output will look like.

Each of the below sections lists the valid options and a description of what the value should be, and then an example of what that section might look like in the INI format:

[hveto]
-------

========================  =====================================================
``snr-thresholds``        A comma-separated list of signal-to-noise ratio
                          thresholds over which to test each auxiliary channel
``time-windows``          A comma-separated list of time windows (in
                          seconds) over which to test each auxiliary channel
``minimum-significance``  The significance value below which to stop the
                          analysis
========================  =====================================================

.. code-block:: ini

   [hveto]
   ; comma-separated list of SNR thresholds (pass >=)
   snr-thresholds = 7.75, 8, 8.5, 9, 10, 11, 12, 15, 20, 40, 100, 300
   ; comma-separated list of time windows
   time-windows = .1, .2, .4, .8, 1
   minimum-significance = 5

[primary]
---------

=====================  ========================================================
``channel``            The name of the primary channel
``trigger-generator``  The name of the primary trigger generator
``snr-threshold``      The minimum threshold on signal-to-noise ratio for
                       primary channel events to be included in the analysis
``frequency-range``    The `(low, high`) frequency range of interest for this
                       analysis. Note that for CBC trigger generators, the
                       ``'template_duration'`` column is used here.
``read-format``        The ``format`` name to use when reading files for this
                       trigger generator
``read-columns``       The list of three column names equivalent to
                       `time-like,frequency-like,snr-like` for this trigger
                       generator
=====================  ========================================================

.. code-block:: ini

   [primary]
   ; use Omicron calibrated strain as primary
   channel = %(IFO)s:GDS-CALIB_STRAIN
   trigger-generator = omicron
   ; SNR threshold (pass >=)
   snr-threshold = 6
   ; flow, fhigh
   frequency-range = 0, 2048.
   ; format
   read-format = ligolw
   read-tablename = sngl_burst
   ; read-columns to read
   read-columns = peak, peak_frequency, snr

.. note::

   In this section the ``%(IFO)s`` interpolation variable has been used. This
   can be used throughout the configuration so that you only have to write one
   for all interferometers, with the correct two-character prefix being
   inserted automatically at run time.

.. note::

   As well as the above, users can specify arguments to be used when
   automatically discovering triggers. Any option that begins ``trigfind-``
   will be passed to the `trigfind.find_trigger_urls` function as a keyword
   argument.

   Specifically, to specify the specific run associated with the ``daily-cbc``
   event generator, you can give

   .. code-block:: ini

      [primary]
      trigger-generator = daily-cbc
      trigfind-run = bbh_gds
      trigfind-filetag = 16SEC_CLUSTERED
      read-format = ligolw
      read-tablename = sngl_inspiral
      read-columns = end,template_duration,snr

[auxiliary]
-----------

=====================  ========================================================
``trigger-generator``  The name of the auxiliary trigger generator
``frequency-range``    The `(low, high)` frequency range of interest
``read-format``        The ``format`` name to use when reading files for this
                       trigger generator
``read-columns``       The list of three column names equivalent to
                       `time-like,frequency-like,snr-like` for this trigger
                       generator
``channels``           a tab-indented, line-delimited list of auxiliary channel
                       names
=====================  ========================================================

.. code-block:: ini

   [auxiliary]
   trigger-generator = omicron
   ; flow, fhigh
   frequency-range = 0, 2048
   ; file format
   read-format = ligolw
   read-tablename = sngl_burst
   ; read-columns to read
   read-columns = peak, peak_frequency, snr
   ; give tab-indented, line-separated list of channels
   channels =
       %(IFO)s:ASC-AS_B_RF45_I_PIT_OUT_DQ
       %(IFO)s:ASC-AS_A_RF45_Q_PIT_OUT_DQ
       %(IFO)s:ASC-X_TR_B_NSUM_OUT_DQ
       %(IFO)s:ISI-ITMX_ST2_BLND_X_GS13_CUR_IN1_DQ
       %(IFO)s:ASC-REFL_A_RF9_I_YAW_OUT_DQ
       %(IFO)s:ASC-AS_A_RF45_Q_YAW_OUT_DQ

.. note::

   As with the ``[primary]`` section, users can specify ``trigfind-``
   arguments to customise trigger-file discovery.

[segments]
----------

===========================  ==================================================
``url``                      The URL of the segment database
``analysis-flag``            The name of the data-quality flag indicating
                             analysable times
``padding``                  The `(pre, post)` padding to apply to the analysis
                             segments [note both `pre` and `post` operate
                             forward in time, so to pad out at the start of a
                             segment (or in at the end), use a negative number]
``veto-definer-file``        The path of a veto-definer file to apply before
                             analysis (can be a remote URL)
``veto-definer-categories``  Comma-separated list of category integers to
                             apply, defaults to all flags in veto-definer file
===========================  ==================================================

.. code-block:: ini

   [segments]
   url = https://segments.ligo.org
   ; require full observation mode
   analysis-flag = %(IFO)s:DMT-ANALYSIS_READY:1
   ; no padding by default
   padding = 0, 0

[safety]
--------

===================  ==========================================================
``unsafe-channels``  The list of unsafe channels (tab-indented, line-delimited)
===================  ==========================================================

.. code-block:: ini

   [safety]
   unsafe-channels =
       %(IFO)s:GDS-CALIB_STRAIN
       %(IFO)s:CAL-DELTAL_EXTERNAL_DQ
       %(IFO)s:OAF-CAL_DARM_DQ
       %(IFO)s:OMC-DCPD_SUM_OUT_DQ

Module API
==========
"""

import os
try:
    import configparser
except ImportError:  # python 2.x
    import ConfigParser as configparser

from .segments import DEFAULT_SEGMENT_SERVER

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Joshua Smith <joshua.smith@ligo.org>'


class HvetoConfigParser(configparser.ConfigParser):
    HVETO_DEFAULTS = {
       'hveto': {
            'snr-thresholds': (8, 10, 12, 15, 20, 50, 100, 300),
            'time-windows': (.1, .2, .5, 1, 2, 5),
            'minimum-significance': 5,
        },
        'segments': {
            'url': DEFAULT_SEGMENT_SERVER,
            'analysis-flag': '%(IFO)s:DMT-ANALYSIS_READY:1',
            'padding': (0, 0),
        },
        'primary': {
            'channel': '%(IFO)s:GDS-CALIB_STRAIN',
            'trigger-generator': 'Omicron',
            'snr-threshold': 8,
            'frequency-range': (30, 2048),
        },
        'auxiliary': {
            'trigger-generator': 'Omicron',
            'frequency-range': (30, 2048),
        },
        'safety': {
            'unsafe-channels': ['%(IFO)s:GDS-CALIB_STRAIN',
                                '%(IFO)s:LDAS-STRAIN',
                                '%(IFO)s:CAL-DELTAL_EXTERNAL_DQ',
                                '%(IFO)s:DER_DATA_H',
                                '%(IFO)s:h_4096Hz',
                                '%(IFO)s:h_16384Hz'],
        }
    }

    def __init__(self, ifo=None, defaults=dict(), **kwargs):
        if ifo is not None:
           defaults.setdefault('IFO', ifo)
        configparser.ConfigParser.__init__(self, defaults=defaults, **kwargs)
        self.set_hveto_defaults()

    def set_hveto_defaults(self):
        for section in self.HVETO_DEFAULTS:
            self.add_section(section)
            for key, val in self.HVETO_DEFAULTS[section].items():
                if key.endswith('channels') and isinstance(val, (tuple, list)):
                    self.set(section, key, '\n'.join(list(val)))
                elif isinstance(val, tuple):
                    self.set(section, key, ', '.join(map(str, val)))
                else:
                    self.set(section, key, str(val))

    def read(self, filenames):
        readok = configparser.ConfigParser.read(self, filenames)
        for f in filenames:
            if f not in readok:
                raise IOError("Cannot read file %r" % f)
        return readok
    read.__doc__ = configparser.ConfigParser.read.__doc__

    def getfloats(self, section, option):
        return self._get(section, comma_separated_floats, option)

    def getparams(self, section, prefix):
        nchar = len(prefix)
        params = dict((key[nchar:], val) for (key, val) in
                      self.items(section) if key.startswith(prefix))
        # try simple typecasting
        for key in params:
            if params[key].lower() in ('true', 'false'):
                params[key] = bool(params[key])
            else:
                try:
                    params[key] = float(params[key])
                except ValueError:
                    pass
        return params


def comma_separated_floats(string):
    return tuple(map(float, string.split(',')))



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the hveto python package.
#
# hveto is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# hveto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with hveto.  If not, see <http://www.gnu.org/licenses/>.

"""General utilities for hveto
"""

from __future__ import (division, print_function)

import sys
import os.path
from math import ceil

from gwdatafind.utils import filename_metadata

try:  # python 3.x
    from io import StringIO
    from html.parser import HTMLParser
except:  # python 2.7
    from cStringIO import StringIO
    from HTMLParser import HTMLParser

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Alex Urban <alexander.urban@ligo.org>'


def write_lal_cache(target, paths):
    # if not an open file, open it
    if isinstance(target, str):
        with open(target, "w") as fobj:
            write_lal_cache(fobj, paths)
            return target

    # write to file
    for path in paths:
        obs, tag, segment = filename_metadata(path)
        print(obs, tag, segment[0], abs(segment), path, file=target)

    return target


# -- class for HTML parsing ---------------------------------------------------

class HvetoHTMLParser(HTMLParser):
    """See https://docs.python.org/3/library/html.parser.html.
    """
    def handle_starttag(self, tag, attrs):
        print("Start tag:", tag)
        attrs.sort()
        for attr in attrs:
            print("attr:", attr)

    def handle_endtag(self, tag):
        print("End tag:", tag)

    def handle_data(self, data):
        print("Data:", data)

    def handle_decl(self, data):
        print("Decl:", data)

parser = HvetoHTMLParser()


# -- utilities ----------------------------------------------------------------

def parse_html(html):
    """Parse a string containing raw HTML code
    """
    stdout = sys.stdout
    sys.stdout = StringIO()
    if sys.version_info.major < 3:
        parser.feed(html.decode('utf-8', 'ignore'))
    else:
        parser.feed(html)
    output = sys.stdout.getvalue()
    sys.stdout = stdout
    return output


def channel_groups(channellist, ngroups):
    """Separate a list of channels into a number of equally-sized groups

    Parameters
    ----------
    channellist : `list`
        large list to separate into chunks
    ngroups : `int`
        number of output groups

    Returns
    -------
    iterator : iterator `list` of `list`
        a generator sequence yielding a sub-list on each iteration
    """
    n = int(ceil(len(channellist) / ngroups))
    for i in range(0, len(channellist), n):
        yield channellist[i:i+n]



# -*- coding: utf-8 -*-
# Copyright (C) Joshua Smith (2016-)
#
# This file is part of the hveto python package.
#
# hveto is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# hveto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with hveto.  If not, see <http://www.gnu.org/licenses/>.

"""HTML utilities for hveto
"""

from __future__ import division

import sys
import os.path
import datetime
from functools import wraps
from getpass import getuser

from pygments import highlight
from pygments.lexers import get_lexer_by_name
from pygments.formatters import HtmlFormatter

from MarkupPy import markup

from gwdetchar.io import html as gwhtml

from ._version import get_versions

__author__ = 'Duncan Macleod <duncan.macleod@ligo.org>'
__credits__ = 'Josh Smith, Joe Areeda, Alex Urban'

# -- set up default JS and CSS files

JQUERY_JS = "//code.jquery.com/jquery-1.11.2.min.js"

BOOTSTRAP_CSS = (
    "//maxcdn.bootstrapcdn.com/bootstrap/3.3.4/css/bootstrap.min.css")
BOOTSTRAP_JS = (
    "//maxcdn.bootstrapcdn.com/bootstrap/3.3.4/js/bootstrap.min.js")

FANCYBOX_CSS = (
    "//cdnjs.cloudflare.com/ajax/libs/fancybox/2.1.5/jquery.fancybox.min.css")
FANCYBOX_JS = (
    "//cdnjs.cloudflare.com/ajax/libs/fancybox/2.1.5/jquery.fancybox.min.js")

CSS_FILES = [BOOTSTRAP_CSS, FANCYBOX_CSS]
JS_FILES = [JQUERY_JS, BOOTSTRAP_JS, FANCYBOX_JS]

HVETO_CSS = """
html {
		position: relative;
		min-height: 100%;
}
body {
		margin-bottom: 120px;
		font-family: "Helvetica Neue", Helvetica, Arial, sans-serif;
		-webkit-font-smoothing: antialiased;
}
.footer {
		position: absolute;
		bottom: 0;
		width: 100%;
		height: 100px;
		background-color: #f5f5f5;
		padding-top: 20px;
		padding-bottom: 20px;
}
.fancybox-skin {
		background: white;
}
"""

HVETO_JS = """
$(document).ready(function() {
	$(\".fancybox\").fancybox({
		nextEffect: 'none',
		prevEffect: 'none',
		helpers: {title: {type: 'inside'}}
	});
});
"""


# -- HTML construction --------------------------------------------------------

def write_static_files(static):
    """Write static CSS and javascript files into the given directory

    Parameters
    ----------
    static : `str`
        the target directory for the static files, will be created if
        necessary

    Returns
    -------
    `<statc>/hveto.css`
    `<static>/hveto.js`
        the paths of the two files written by this method, which will be
        `hveto.css` and `hveto.js` inside `static`

    Notes
    -----
    This method modifies the module-level variables ``CSS_FILES` and
    ``JS_FILES`` to guarantee that the static files are actually only
    written once per process.
    """
    if not os.path.isdir(static):
        os.makedirs(static)
    hvetocss = os.path.join(static, 'hveto.css')
    if hvetocss not in CSS_FILES:
        with open(hvetocss, 'w') as f:
            f.write(HVETO_CSS)
        CSS_FILES.append(hvetocss)
    hvetojs = os.path.join(static, 'hveto.js')
    if hvetojs not in JS_FILES:
        with open(hvetojs, 'w') as f:
            f.write(HVETO_JS)
        JS_FILES.append(hvetojs)
    return hvetocss, hvetojs


def init_page(ifo, start, end, css=[], script=[], base=os.path.curdir,
              **kwargs):
    """Initialise a new `markup.page`

    This method constructs an HTML page with the following structure

    .. code-block:: html

        <html>
        <head>
            <!-- some stuff -->
        </head>
        <body>
        <div class="container">
        <div class="page-header">
        <h1>IFO</h1>
        <h3>START-END</h3>
        </div>
        </div>
        <div class="container">

    Parameters
    ----------
    ifo : `str`
        the interferometer prefix
    start : `int`
        the GPS start time of the analysis
    end : `int`
        the GPS end time of the analysis
    css : `list`, optional
        the list of stylesheets to link in the `<head>`
    script : `list`, optional
        the list of javascript files to link in the `<head>`
    base : `str`, optional, default '.'
        the path for the `<base>` tag to link in the `<head>`

    Returns
    -------
    page : `markup.page`
        the structured markup to open an HTML document
    """
    # write CSS to static dir
    staticdir = os.path.join(os.path.curdir, 'static')
    write_static_files(staticdir)
    # create page
    page = markup.page()
    page.header.append('<!DOCTYPE HTML>')
    page.html(lang='en')
    page.head()
    page.base(href=base)
    page._full = True
    # link stylesheets (appending bootstrap if needed)
    css = css[:]
    for cssf in CSS_FILES[::-1]:
        b = os.path.basename(cssf)
        if not any(f.endswith(b) for f in css):
            css.insert(0, cssf)
    for f in css:
        page.link(href=f, rel='stylesheet', type='text/css', media='all')
    # link javascript
    script = script[:]
    for jsf in JS_FILES[::-1]:
        b = os.path.basename(jsf)
        if not any(f.endswith(b) for f in script):
            script.insert(0, jsf)
    for f in script:
        page.script('', src=f, type='text/javascript')
    # add other attributes
    for key in kwargs:
        getattr(page, key)(kwargs[key])
    # finalize header
    page.head.close()
    page.body()
    # write banner
    page.div(class_='container')
    page.div(class_='page-header', role='banner')
    page.h1("%s HierarchicalVeto" % ifo)
    page.h3("%d-%d" % (start, end))
    page.div.close()
    page.div.close()  # container

    # open container
    page.div(class_='container')
    return page


def close_page(page, target, about=None, date=None):
    """Close an HTML document with markup then write to disk

    This method writes the closing markup to complement the opening
    written by `init_page`, something like:

    .. code-block:: html

       </div>
       <footer>
           <!-- some stuff -->
       </footer>
       </body>
       </html>

    Parameters
    ----------
    page : `markup.page`
        the markup object to close
    target : `str`
        the output filename for HTML
    about : `str`, optional
        the path of the 'about' page to link in the footer
    date : `datetime.datetime`, optional
        the timestamp to place in the footer, defaults to
        `~datetime.datetime.now`
    """
    page.div.close()  # container
    page.add(str(write_footer(about=about, date=date)))
    if not page._full:
        page.body.close()
        page.html.close()
    with open(target, 'w') as f:
        f.write(page())
    return page


def wrap_html(func):
    """Decorator to wrap a function with `init_page` and `close_page` calls

    This allows inner HTML methods to be written with minimal arguments
    and content, hopefully making things simpler
    """
    @wraps(func)
    def decorated_func(ifo, start, end, *args, **kwargs):
        # set page init args
        initargs = {
            'title': '%s Hveto | %d-%d' % (ifo, start, end),
            'base': os.path.curdir,
        }
        for key in ['title', 'base']:
            if key in kwargs:
                initargs[key] = kwargs.pop(key)
        # find outdir
        outdir = kwargs.pop('outdir', initargs['base'])
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        # write about page
        try:
            config = kwargs.pop('config')
        except KeyError:
            about = None
        else:
            iargs = initargs.copy()
            aboutdir = os.path.join(outdir, 'about')
            if iargs['base'] == os.path.curdir:
                iargs['base'] = os.path.pardir
            about = write_about_page(ifo, start, end, config, outdir=aboutdir,
                                     **iargs)
            if os.path.basename(about) == 'index.html':
                about = about[:-10]
        # open page
        page = init_page(ifo, start, end, **initargs)
        # write content
        contentf = os.path.join(outdir, '_inner.html')
        with open(contentf, 'w') as f:
            f.write(str(func(*args, **kwargs)))
        # embed content
        page.div('', id_='content')
        page.script("$('#content').load('%s');" % contentf)
        # close page
        index = os.path.join(outdir, 'index.html')
        close_page(page, index, about=about)
        return index
    return decorated_func


# -- Utilities ----------------------------------------------------------------

def bold_param(key, value, **attrs):
    """Write a (key, value) pair in HTML as ``<p><b>key</b>: value</p>``

    Parameters
    ----------
    key : `str`
        the element to be rendered in bold
    value : `str`
        the explanation/description of the key
    **attrs
        HTML attributes for the containing ``<p>`` tag

    Returns
    -------
    html : `str`
    """
    return markup.oneliner.p('<b>%s</b>: %s' % (key, value), **attrs)


def write_footer(about=None, date=None):
    """Write a <footer> for an Hveto page

    Parameters
    ----------
    about : `str`, optional
        path of about page to link
    date : `datetime.datetime`, optional
        the datetime representing when this analysis was generated, defaults
        to `~datetime.datetime.now`

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the markup object containing the footer HTML
    """
    page = markup.page()
    page.twotags.append('footer')
    markup.element('footer', case=page.case, parent=page)(class_='footer')
    page.div(class_='container')
    # write user/time for analysis
    if date is None:
        date = datetime.datetime.now().replace(second=0, microsecond=0)
    version = get_versions()['version']
    commit = get_versions()['full-revisionid']
    url = 'https://github.com/gwdetchar/hveto/tree/%s' % commit
    link = markup.oneliner.a('hveto version %s' % version, href=url,
                             target='_blank')
    page.p('These results were obtained using {link} by {user} at '
           '{date}.'.format(link=link, user=getuser(), date=date))
    # link to 'about'
    if about is not None:
        page.a('How was this page generated?', href=about)
    page.div.close()  # container
    markup.element('footer', case=page.case, parent=page).close()
    return page


# -- Hveto HTML ---------------------------------------------------------------

def write_summary(
        rounds, plots=[], header='Summary', plotsperrow=4,
        tableclass='table table-condensed table-hover table-responsive'):
    """Write the Hveto analysis summary HTML

    Parameters
    ----------
    rounds : `list` of `HvetoRound`
        the `list` of round objects produced by this analysis
    plots : `list` of `str`, optional
        the `list` of summary plots to display underneath the summary table
    header : `str`, optional
        the text for the section header (``<h2``>)
    plotsperrow : `int`, optional
        the number of plots to display in each row
    tableclass : `str`, optional
        the ``class`` for the summary ``<table>``

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the formatted markup object containing the analysis summary table,
        and images
    """
    page = markup.page()
    page.h2(header)
    page.table(class_=tableclass)
    page.caption("Summary of this HierarchichalVeto analysis.")
    # make header
    page.thead()
    page.tr()
    for header in ['Round', 'Winner', 'Twin [s]', 'SNR Thresh', 'Significance',
                   'Use [%]', 'Efficiency [%]', 'Deadtime [%]',
                   'Cum. efficiency [%]', 'Cum. deadtime [%]']:
        page.th(header, scope='row')
    page.tr.close()
    page.thead.close()
    # make body
    page.tbody()
    for r in rounds:
        page.tr()
        # link round down page
        page.td(gwhtml.html_link('#hveto-round-%d' % r.n, r.n, target=None,
                                 title="Jump to round %d details" % r.n))
        # link name to CIS
        page.td(gwhtml.cis_link(r.winner.name))
        for attr in ['window', 'snr', 'significance']:
            v = getattr(r.winner, attr)
            if isinstance(v, float):
                page.td('%.2f' % v)
            else:
                page.td(str(v))
        for attr in ['use_percentage', 'efficiency', 'deadtime',
                     'cum_efficiency', 'cum_deadtime']:
            a, b = getattr(r, attr)
            try:
                pc = a/b * 100.
            except ZeroDivisionError:
                pc = 0.
            if attr.endswith('deadtime'):
                page.td('%.2f<br><small>[%.2f/%.2f]</small>' % (pc, a, b))
            else:
                page.td('%.2f<br><small>[%d/%d]</small>' % (pc, a, b))
        page.tr.close()
    page.tbody.close()

    page.table.close()

    # scaffold plots
    if plots:
        page.add(gwhtml.scaffold_plots(plots, nperrow=plotsperrow))
    return page()


def write_round(round):
    """Write the HTML summary for a specific round

    Parameters
    ----------
    round : `HvetoRound`
        the analysis round object

    Returns
    -------
    page : `~MarkupPy.markup.page`
        the formatted HTML for this round
    """
    page = markup.page()
    page.div(class_='panel panel-info', id_='hveto-round-%d' % round.n)
    # -- make heading
    page.div(class_='panel-heading clearfix')
    # link to top of page
    page.div(class_='pull-right')
    page.a("<small>[top]</small>", href='#')
    page.div.close()  # pull-right
    # heading
    page.h3('Round %d, Winner = %s, window = %s, SNR thresh = %s'
            % (round.n, round.winner.name, round.winner.window,
               round.winner.snr), class_='panel-title')
    page.div.close()  # panel-heading

    # -- make body
    page.div(class_='panel-body')
    page.div(class_='row')
    # summary information
    page.div(class_='col-md-3', id_='hveto-round-%d-summary' % round.n)
    page.add(bold_param('Winner', round.winner.name))
    page.add(bold_param('SNR threshold', round.winner.snr))
    page.add(bold_param('Window', round.winner.window))
    page.add(bold_param('Significance', '%.2f' % round.winner.significance))
    for desc, tag in zip(
            ['Veto segments', 'Veto triggers', 'Vetoed primary triggers',
             'Unvetoed primary triggers'],
            ['VETO_SEGS', 'WINNER', 'VETOED', 'RAW']):
        if isinstance(round.files[tag], str):
            files = [round.files[tag]]
        else:
            files = round.files[tag]
        link = ' '.join([gwhtml.html_link(
            f, '[%s]' % os.path.splitext(f)[1].strip('.')) for f in files])
        page.add(bold_param(desc, link))
    # link omega scans if generated
    if round.scans is not None:
        page.p('<b>Omega scans:</b>')
        for t in round.scans:
            page.p()
            page.a('%s [SNR %.1f]' % (t['time'], t['snr']),
                href='./scans/%s/' % t['time'], **{
                'class_': 'fancybox',
                'data-fancybox-group': 'hveto-image',
                'target': '_blank',
            })
            for c, tag in zip([round.primary, round.winner.name],
                              ['Primary', 'Auxiliary']):
                caption = 'Omega scan of %s at %s' % (c, t['time'])
                png = ('./scans/%s/plots/%s-qscan_whitened-1.png'
                       % (t['time'], c.replace('-', '_').replace(':', '-')))
                page.a('[%s]' % tag[0].lower(), class_='fancybox',
                       href=png, title=caption,
                       **{'data-fancybox-group': 'omega-preview'})
            page.p.close()
    page.div.close()  # col
    # plots
    page.div(class_='col-md-9', id_='hveto-round-%d-plots' % round.n)
    page.add(gwhtml.scaffold_plots(round.plots[:-1], nperrow=4))
    # add significance drop plot at end
    page.div(class_='row')
    page.div(class_='col-sm-12')
    page.add(gwhtml.fancybox_img(round.plots[-1]))
    page.div.close()  # col-sm-12
    page.div.close()  # row
    page.div.close()  # col-md-8

    page.div.close()  # row
    # close and return
    page.div.close()  # panel-body
    page.div.close()  # panel
    return page()


# reminder: wrap_html automatically prepends the (ifo, start, end) args,
# and at least the outdir kwarg, so you should include those in the docstring,
# but not in the actual function declaration - the decorator will take care of
# that for you.

@wrap_html
def write_hveto_page(rounds, plots):
    """Write the Hveto results to HTML

    Parameters
    ----------
    ifo : `str`
        the prefix of the interferometer used in this analysis
    start  : `int`
        the GPS start time of the analysis
    end : `int`
        the GPS end time of the analysis
    rounds : `list` of `HvetoRound`
        the rounds produced by this analysis
    plots : `list` of `str`
        the `list` of summary plots
    outdir : `str`, optional
        the output directory for the HTML

    Returns
    -------
    index : `str`
        the path of the HTML written for this analysis
    """
    page = markup.page()
    page.add(write_summary(rounds, plots))
    page.h2('Round details')
    for r in rounds:
        page.add(write_round(r))
    return page


@wrap_html
def write_null_page(reason, context='info'):
    """Write the Hveto results to HTML

    Parameters
    ----------
    ifo : `str`
        the prefix of the interferometer used in this analysis
    start  : `int`
        the GPS start time of the analysis
    end : `int`
        the GPS end time of the analysis
    reason : `str`
        the explanation for this null result
    context : `str`, optional
        the bootstrap context class for this result, see the bootstrap
        docs for more details
    outdir : `str`, optional
        the output directory for the HTML

    Returns
    -------
    index : `str`
        the path of the HTML written for this analysis
    """
    page = markup.page()
    # write alert
    page.div(class_='alert alert-%s' % context)
    page.p(reason)
    page.div.close()  # alert
    return page


@wrap_html
def write_about_page(configfile):
    """Write a page explaining how an hveto analysis was completed

    Parameters
    ----------
    ifo : `str`
        the prefix of the interferometer used in this analysis
    start  : `int`
        the GPS start time of the analysis
    end : `int`
        the GPS end time of the analysis
    configfile : `str`
        the path of the configuration file to embed
    outdir : `str`, optional
        the output directory for the HTML

    Returns
    -------
    index : `str`
        the path of the HTML written for this analysis
    """
    # configure syntax highlighting
    blexer = get_lexer_by_name('bash', stripall=True)
    ilexer = get_lexer_by_name('ini', stripall=True)
    formatter = HtmlFormatter(noclasses=True)
    # set up page
    page = markup.page()

    # command line
    page.h2('On the command line')
    page.p('This page was generated with the command line call shown below.')
    commandline = highlight(' '.join(sys.argv), blexer, formatter)
    page.add(commandline)

    # configuration file
    page.h2('Configuration')
    with open(configfile, 'r') as fobj:
        inifile = fobj.read()
    contents = highlight(inifile, ilexer, formatter)
    page.add(contents)

    # runtime environment
    page.add(gwhtml.package_table())

    return page
